/*
 * Copyright 2010 Canonical, Ltd.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Authors:
 *    Cody Russell  <crussell@canonical.com>
 */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>

#include <dbus/dbus.h>
#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-lowlevel.h>
#include <dbus/dbus-glib-bindings.h>

#include <gtk/gtk.h>
#include <gdk/gdkx.h>

#include <libdbusmenu-glib/menuitem.h>
#include <libdbusmenu-glib/server.h>

#include "bridge.h"
#include "application-menu-registrar-client.h"

#define APP_MENU_DBUS_NAME   "org.ayatana.WindowMenu.Registrar"
#define APP_MENU_DBUS_OBJECT "/org/ayatana/WindowMenu/Registrar"
#define APP_MENU_INTERFACE   "org.ayatana.WindowMenu.Registrar"
#define APP_MENU_PATH        "/this/is/a/long/object/path"

static void app_menu_bridge_insert         (GtkMenuProxy      *proxy,
                                            GtkWidget         *shell,
                                            GtkWidget         *child,
                                            guint              position);

struct _AppMenuBridgePrivate
{
  GHashTable *items;      /* <GtkWidget *, DbusmenuMenuitem *> */
  GHashTable *unparented; /* <GtkWidget *, DbusmenuMenuitem *> */

  DbusmenuServer   *server;
  DbusmenuMenuitem *root;
};

static DBusGProxy *dbusproxy = NULL;

G_DEFINE_DYNAMIC_TYPE(AppMenuBridge, app_menu_bridge, GTK_TYPE_MENU_PROXY)

static void
app_menu_bridge_init (AppMenuBridge *bridge)
{
  bridge->priv = G_TYPE_INSTANCE_GET_PRIVATE (bridge, APP_MENU_TYPE_BRIDGE, AppMenuBridgePrivate);

  bridge->priv->items = g_hash_table_new (g_direct_hash, g_direct_equal);
  bridge->priv->items = g_hash_table_new (g_direct_hash, g_direct_equal);

  bridge->priv->server = dbusmenu_server_new (APP_MENU_PATH);
  bridge->priv->root = dbusmenu_menuitem_new ();

  dbusmenu_server_set_root (bridge->priv->server,
                            bridge->priv->root);
}

static void
app_menu_bridge_finalize (GObject *object)
{
  g_hash_table_destroy (APP_MENU_BRIDGE (object)->priv->items);
}

static void
app_menu_bridge_class_init (AppMenuBridgeClass *class)
{
  GtkMenuProxyClass *proxy_class;
  GObjectClass *object_class;

  app_menu_bridge_parent_class = g_type_class_peek_parent (class);

  proxy_class = GTK_MENU_PROXY_CLASS (class);
  object_class = G_OBJECT_CLASS (class);

  proxy_class->insert = app_menu_bridge_insert;

  object_class->finalize = app_menu_bridge_finalize;

  g_type_class_add_private (class, sizeof (AppMenuBridgePrivate));
}

static void
app_menu_bridge_class_finalize (AppMenuBridgeClass *class)
{
}

static GtkWidget *
find_menu_label (GtkWidget *widget)
{
  GtkWidget *label = NULL;

  if (GTK_IS_LABEL (widget))
    return widget;

  if (GTK_IS_CONTAINER (widget))
    {
      GList *children;
      GList *l;

      children = gtk_container_get_children (GTK_CONTAINER (widget));

      for (l = children; l; l = l->next)
        {
          label = find_menu_label (l->data);

          if (label)
            break;
        }

      g_list_free (children);
    }

  return label;
}

static const gchar *
get_menu_label_text (GtkWidget *menuitem)
{
  GtkWidget *label = find_menu_label (menuitem);

  if (label)
    return gtk_label_get_text (GTK_LABEL (label));

  return NULL;
}

static void
item_activated (DbusmenuMenuitem *item, guint timestamp, gpointer user_data)
{
  GtkWidget *child;

  if (user_data != NULL)
    {
      child = (GtkWidget *)user_data;

      if (GTK_IS_MENU_ITEM (child))
        {
          gtk_menu_item_activate (GTK_MENU_ITEM (child));
        }
    }
}

static void
widget_notify_cb (GtkWidget  *widget,
                  GParamSpec *pspec,
                  gpointer    data)
{
  DbusmenuMenuitem *child = (DbusmenuMenuitem *)data;

  if (pspec->name == g_intern_static_string ("sensitive"))
    {
      dbusmenu_menuitem_property_set_bool (child,
                                           DBUSMENU_MENUITEM_PROP_ENABLED,
                                           GTK_WIDGET_IS_SENSITIVE (widget));
    }
  else if (pspec->name == g_intern_static_string ("label"))
    {
      dbusmenu_menuitem_property_set (child,
                                      DBUSMENU_MENUITEM_PROP_LABEL,
                                      gtk_menu_item_get_label (GTK_MENU_ITEM (widget)));
    }
  else if (pspec->name == g_intern_static_string ("visible"))
    {
      dbusmenu_menuitem_property_set_bool (child,
                                           DBUSMENU_MENUITEM_PROP_VISIBLE,
                                           gtk_widget_get_visible (widget));
    }
  else if (pspec->name == g_intern_static_string ("parent-set"))
    {
      g_print (" ** parent set!!\n");
    }
}

static void
toplevel_realized (GtkWidget *widget,
                   gpointer   user_data)
{
  org_ayatana_WindowMenu_Registrar_register_window (dbusproxy,
                                                    GDK_WINDOW_XID (gtk_widget_get_window (widget)),
                                                    APP_MENU_PATH,
                                                    NULL);
}

static void
app_menu_bridge_insert (GtkMenuProxy *proxy,
                        GtkWidget    *parent,
                        GtkWidget    *child,
                        guint         position)
{
  static gboolean   window_registered = FALSE;
  AppMenuBridge    *bridge;
  DbusmenuMenuitem *item;
  GtkWidget        *submenu;
  DbusmenuMenuitem *parent_item = NULL;

  if (GTK_IS_TEAROFF_MENU_ITEM (child) || GTK_IS_MENU (child))
    return;

  if (!gtk_widget_get_visible (child))
    return;

  if (!window_registered && GTK_IS_MENU_BAR (parent))
    {
      GtkWidget *toplevel = gtk_widget_get_toplevel (parent);

      g_signal_connect (toplevel, "realize",
                        G_CALLBACK (toplevel_realized),
                        NULL);
    }

  bridge = APP_MENU_BRIDGE (proxy);

  if (GTK_IS_MENU_BAR (parent))
    {
      if (!g_hash_table_lookup (bridge->priv->items, parent))
        {
          g_hash_table_insert (bridge->priv->items, parent, bridge->priv->root);
        }

      parent_item = bridge->priv->root;
    }
  else if (GTK_IS_MENU (parent))
    {
      GtkWidget *attached_to;

      g_object_get (parent,
                    "attach-widget", &attached_to,
                    NULL);

      if (attached_to)
        {
          parent_item = g_hash_table_lookup (bridge->priv->items, attached_to);
        }
    }

  if (GTK_IS_MENU_ITEM (child))
    {
      item = dbusmenu_menuitem_new ();
      g_hash_table_insert (bridge->priv->items, child, item); // move to end?

      if (GTK_IS_SEPARATOR_MENU_ITEM (child))
        {
          dbusmenu_menuitem_property_set (item,
                                          "type",
                                          "separator");
        }
      else
        {
          dbusmenu_menuitem_property_set (item,
                                          "label",
                                          get_menu_label_text (child));

          if (!gtk_menu_item_get_submenu (GTK_MENU_ITEM (child)))
            {
              g_signal_connect (G_OBJECT (item),
                                "item_activated",
                                G_CALLBACK (item_activated),
                                child);
            }
        }

      g_signal_connect (G_OBJECT (child),
                        "notify",
                        G_CALLBACK (widget_notify_cb),
                        item);

      submenu = gtk_menu_item_get_submenu (GTK_MENU_ITEM (child));
      if (submenu != NULL && !g_hash_table_lookup (bridge->priv->items, submenu))
        {
          g_hash_table_insert (bridge->priv->items, submenu, item);
        }

      if (parent == NULL)
        {
          parent_item = bridge->priv->root;
        }

      if (parent_item)
        {
          /* XXX - not sure about this. :)   */
          if (GTK_IS_MENU_BAR (parent))
            dbusmenu_menuitem_child_append (parent_item, item);
          else
            dbusmenu_menuitem_child_prepend (parent_item, item);
        }
    }
}

G_MODULE_EXPORT void
menu_proxy_module_load (GtkMenuProxyModule *module)
{
  static gboolean registered = FALSE;

  if (!registered)
    {
      DBusGConnection *connection;

      connection = dbus_g_bus_get (DBUS_BUS_SESSION, NULL);

      g_return_if_fail (connection != NULL);

      dbusproxy = dbus_g_proxy_new_for_name_owner (connection,
                                                   APP_MENU_DBUS_NAME,
                                                   APP_MENU_DBUS_OBJECT,
                                                   APP_MENU_INTERFACE,
                                                   NULL);

      g_return_if_fail (dbusproxy != NULL);

      app_menu_bridge_register_type (G_TYPE_MODULE (module));

      registered = TRUE;
    }
}

G_MODULE_EXPORT void
menu_proxy_module_unload (GtkMenuProxyModule *module)
{
}
